#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <boost/config.hpp>
#include <boost/graph/graph_traits.hpp>
#include <boost/graph/graph_concepts.hpp>
#include <boost/graph/properties.hpp>
#include <boost/graph/named_function_params.hpp>
#include <boost/ref.hpp>
#include <boost/implicit_cast.hpp>
#include <boost/graph/filtered_graph.hpp>
#include <boost/graph/strong_components.hpp>
#include <boost/graph/topological_sort.hpp>
#include <vector>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/graph_utility.hpp>
#include <iostream>
#include <string>
#include <sstream>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/properties.hpp>
#include <boost/graph/graphviz.hpp>
#include <boost/property_map/property_map.hpp>
#include <queue>
#include <ilcplex/ilocplex.h>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/breadth_first_search.hpp>
#include <boost/graph/floyd_warshall_shortest.hpp>
#include <boost/graph/dijkstra_shortest_paths.hpp>
#include <boost/graph/dijkstra_shortest_paths_no_color_map.hpp>
#include <boost/graph/visitors.hpp>
#include <boost/graph/copy.hpp>
#include <boost/graph/reverse_graph.hpp> 
#include <limits.h>

using namespace std;
using namespace boost;

// User defined propety
enum vertex_second_name_t { vertex_second_name };
enum vertex_filtered_t { vertex_filtered };
enum edge_shortest_paths_counter_t { edge_shortest_paths_counter };
enum edge_confidence_t { edge_confidence };
enum edge_treshhold_t { edge_treshhold };
enum edge_in_shortest_t { edge_in_shortest };
enum edge_direction_t { edge_direction };
enum edge_hidden_t { edge_hidden };
enum edge_var_t {edge_var};
enum edge_constraint_t {edge_constraint};
enum edge_flow2_t {edge_flow2};
enum edge_test_t { edge_test };
enum edge_orientation_round_t { edge_orientation_round };
enum edge_target_distances_t { edge_target_distances };
enum edge_ilp_improve_t { edge_ilp_improve };

namespace boost {
	BOOST_INSTALL_PROPERTY(vertex, second_name);
	BOOST_INSTALL_PROPERTY(vertex, filtered);
	BOOST_INSTALL_PROPERTY(edge, shortest_paths_counter);
	BOOST_INSTALL_PROPERTY(edge, in_shortest);
	BOOST_INSTALL_PROPERTY(edge, var);
	BOOST_INSTALL_PROPERTY(edge, constraint);
	BOOST_INSTALL_PROPERTY(edge, flow2);
	BOOST_INSTALL_PROPERTY(edge, direction);
	BOOST_INSTALL_PROPERTY(edge, hidden);
	BOOST_INSTALL_PROPERTY(edge, test);
	BOOST_INSTALL_PROPERTY(edge, orientation_round);
	BOOST_INSTALL_PROPERTY(edge, target_distances);
	BOOST_INSTALL_PROPERTY(edge, confidence);
	BOOST_INSTALL_PROPERTY(edge, treshhold);
	BOOST_INSTALL_PROPERTY(edge, ilp_improve);

	template < typename TimeMap > class bfs_time_visitor:public default_bfs_visitor {
	typedef typename property_traits < TimeMap >::value_type T;
public:
	bfs_time_visitor(TimeMap tmap, T & t):m_timemap(tmap), m_time(t) { }
	template < typename Vertex, typename Graph >
	void discover_vertex(Vertex u, const Graph & g) const
	{
		put(m_timemap, u, m_time++);
	}
  TimeMap m_timemap;
  T & m_time;
	};
				
};



// utility class
template<class P, int N>
struct get_prop_type
{
	typedef typename get_prop_type<typename P::next_type, N-1>::type type;
};

template<class P>
struct get_prop_type<P,0>
{
	typedef P type;
};

// Basic graph: bidirecteded
typedef	property<vertex_name_t, string,
		property<vertex_second_name_t, string,
		property<vertex_distance_t, int, 
		 property<vertex_index_t, int,
		property<vertex_filtered_t, bool,
		property<vertex_discover_time_t, std::size_t,  
		property<vertex_color_t, boost::default_color_type  
		 > > > > > > > vertex_properties;


// Properties available on edges
typedef property<edge_weight_t, int,
		property<edge_confidence_t, double,
		property<edge_treshhold_t, double,
		property<edge_shortest_paths_counter_t, int,
		property<edge_in_shortest_t, bool,
		property<edge_var_t, IloNumVar,
		property<edge_constraint_t, IloConstraint,
		property<edge_flow2_t, IloNumVar,
		property<edge_flow_t, IloNumVar,
		property<edge_direction_t, int,
		property<edge_hidden_t, int,
		property<edge_test_t, int ,
		property<edge_orientation_round_t, int ,
		property<edge_target_distances_t, int ,
		property<edge_index_t, int ,
		property<edge_ilp_improve_t, bool > > > > > > > > > > > > > > > > edge_properties; 

typedef adjacency_list<
		listS,
		listS,
		bidirectionalS,
		vertex_properties,
		edge_properties
		> bidirected_basic_graph;

// graph_traits<> - handles on both vertices and edges
typedef graph_traits<bidirected_basic_graph>::vertex_descriptor  vertex_descriptor;
typedef graph_traits<bidirected_basic_graph>::vertex_iterator    vertex_iterator;
typedef graph_traits<bidirected_basic_graph>::edge_descriptor    edge_descriptor;
typedef graph_traits<bidirected_basic_graph>::edge_iterator      edge_iterator;
typedef graph_traits<bidirected_basic_graph>::out_edge_iterator  out_edge_iterator;
typedef graph_traits<bidirected_basic_graph>::in_edge_iterator   in_edge_iterator;

typedef property_map<bidirected_basic_graph, std::size_t vertex_properties::*>::type dtime_map_t;

// property_map<> - access the various properties on either edges on nodes
typedef property_map<bidirected_basic_graph, vertex_discover_time_t>::type       discover_time_map_t;
typedef property_map<bidirected_basic_graph, vertex_color_t>::type       color_t;
typedef property_map<bidirected_basic_graph, vertex_distance_t>::type       distance_t;
typedef property_map<bidirected_basic_graph, vertex_name_t>::type       names_map_t;
typedef property_map<bidirected_basic_graph, edge_confidence_t>::type       confidence_map_t;
typedef property_map<bidirected_basic_graph, vertex_filtered_t>::type       filtered_map_t;
typedef property_map<bidirected_basic_graph, vertex_index_t>::type vertex_myindex_map_t;
typedef property_map<bidirected_basic_graph, edge_ilp_improve_t>::type ilp_improve_map_t;

typedef property_map<bidirected_basic_graph, edge_index_t>::type    myindex_map_t;
typedef property_map<bidirected_basic_graph, edge_shortest_paths_counter_t>::type    shortest_paths_counter_map_t;
typedef property_map<bidirected_basic_graph, edge_in_shortest_t>::type    in_shortest_map_t;
typedef property_map<bidirected_basic_graph, edge_var_t>::type    vars_map_t;
typedef property_map<bidirected_basic_graph, edge_constraint_t>::type    constraints_map_t;
typedef property_map<bidirected_basic_graph, edge_flow2_t>::type    flow2s_map_t;
typedef property_map<bidirected_basic_graph, edge_direction_t>::type    directions_map_t;
typedef property_map<bidirected_basic_graph, edge_test_t>::type    test_map_t;
typedef property_map<bidirected_basic_graph, edge_hidden_t>::type    hidden_map_t;
typedef property_map<bidirected_basic_graph, edge_orientation_round_t>::type    orientation_round_map_t;
typedef property_map<bidirected_basic_graph, edge_target_distances_t>::type    target_distances_map_t;
typedef property_map<bidirected_basic_graph, edge_treshhold_t>::type    treshhold_map_t;


// Graphs' instances
bidirected_basic_graph bi_graph;
bidirected_basic_graph solution_graph;
bidirected_basic_graph copy_bi_graph;
bidirected_basic_graph comp_graph;
bidirected_basic_graph unconfident_solution_comp_graph;
bidirected_basic_graph influence_graph;

// Comp_graph
typedef iterator_property_map<std::vector<vertex_descriptor>::iterator,vertex_myindex_map_t,vertex_descriptor,vertex_descriptor&> IsoMap;


// Get the different properties map associated with the edges
directions_map_t comp_directions_map = get( edge_direction, comp_graph );
target_distances_map_t comp_target_distances = get(edge_target_distances, comp_graph);
discover_time_map_t dtime_comp_map = get(vertex_discover_time, comp_graph );
names_map_t comp_names_map = get( vertex_name, comp_graph );
distance_t comp_distances_map = get( vertex_distance, comp_graph );
distance_t unconfident_sol_comp_distances_map = get( vertex_distance, comp_graph );
vertex_myindex_map_t vertex_comp_myindex_map = get(vertex_index, comp_graph);
vertex_myindex_map_t unconfident_sol_comp_index_map = get(vertex_index, comp_graph);

// Get the different properties map associated with the vertices
discover_time_map_t dtime_map = get(vertex_discover_time, bi_graph );
names_map_t names_map = get( vertex_name, bi_graph );
color_t colors_map = get( vertex_color, bi_graph );
distance_t distances_map = get( vertex_distance, bi_graph );
names_map_t sol_names_map = get( vertex_name, solution_graph );
filtered_map_t filtered_map = get( vertex_filtered, solution_graph );
treshhold_map_t treshhold_map = get( edge_treshhold, solution_graph );
ilp_improve_map_t ilp_improve_map = get( edge_ilp_improve, bi_graph );

// Get the different properties map associated with the edges
property_map<bidirected_basic_graph, edge_weight_t>::type weightmap = get(edge_weight, bi_graph);
confidence_map_t parallel_confidence_map = get(edge_confidence, bi_graph);
confidence_map_t sol_confidence_map = get(edge_confidence, solution_graph);
myindex_map_t myindex_map = get( edge_index, bi_graph );
vertex_myindex_map_t sol_index_map = get( vertex_index, solution_graph );
shortest_paths_counter_map_t shortest_paths_counter_map = get( edge_shortest_paths_counter, bi_graph );
in_shortest_map_t in_shortest_map = get( edge_in_shortest, bi_graph );
vars_map_t vars_map = get( edge_var, bi_graph );
constraints_map_t constraints_map = get( edge_constraint, bi_graph );
vars_map_t sol_vars_map = get( edge_var, solution_graph );
flow2s_map_t flows_map = get( edge_flow2, bi_graph );
directions_map_t directions_map = get( edge_direction, bi_graph );
test_map_t test_map = get( edge_test, bi_graph );
test_map_t sol_test_map = get( edge_test, solution_graph );
hidden_map_t hidden_map = get( edge_hidden, bi_graph );
vertex_myindex_map_t vertex_myindex_map = get(vertex_index, bi_graph);
orientation_round_map_t orientation_round_map = get(edge_orientation_round, solution_graph);
target_distances_map_t target_distances = get(edge_target_distances, bi_graph);


template <typename EdgeInShortestMap>
struct shortest_graph 
{
  shortest_graph() { }
  shortest_graph(EdgeInShortestMap in_shortest) : m_in_shortest(in_shortest) { }
  template <typename Edge>
  bool operator()(const Edge& e) const 
  {
	return (in_shortest_map[e] == true);
  }
  EdgeInShortestMap m_in_shortest;
};

shortest_graph<in_shortest_map_t> filter_in_shortest(in_shortest_map);
filtered_graph<bidirected_basic_graph, shortest_graph<in_shortest_map_t> > in_shortest_filtered(bi_graph, filter_in_shortest);

// Define graph traits on fiters
typedef graph_traits<filtered_graph<bidirected_basic_graph, shortest_graph<in_shortest_map_t> > > ::edge_iterator in_shortest_filter_edge_iterator;
typedef graph_traits<filtered_graph<bidirected_basic_graph, shortest_graph<in_shortest_map_t> > > ::in_edge_iterator in_shortest_filter_in_edge_iterator;
typedef graph_traits<filtered_graph<bidirected_basic_graph, shortest_graph<in_shortest_map_t> > > ::vertex_iterator in_shortest_filter_vertex_iterator;
typedef graph_traits<filtered_graph<bidirected_basic_graph, shortest_graph<in_shortest_map_t> > > ::out_edge_iterator in_shortest_filter_out_edge_iterator;

// For Boost:: BFS purposes
typedef graph_traits<bidirected_basic_graph>::vertices_size_type Size;
typedef Size* Iiter;

// BFS's Queue
//std::vector<vertex_descriptor> bfs_queue;
int num_of_edges_redirected = 0;
// Collection P of pairs
const int MAX_PROTEIN_ID = 10000;

struct protein_index
{
	vertex_descriptor protein;
	vertex_descriptor protein_tree;
	vertex_descriptor protein_cycle;
};

struct solution_edge_data
{
	vertex_descriptor bi_source;
	vertex_descriptor bi_target;
	double treshhold_map;
	double sol_confidence_map;
	int shortest_paths_counter_map;
	int hidden_map;
	int orientation_round;
	int test_map;
	int directions_map;
	IloNumVar var;

};

struct save_edge
{
	vertex_descriptor source_vertex;
	vertex_descriptor target_vertex;
	IloNumVar save_edge_variable;
	int save_test_map;
	int save_hidden_map;
	int save_shortest_paths_counter_map;
	double save_confidence_map;
	double save_direction;
};

vector<solution_edge_data> solution_edges;
vector<solution_edge_data> unconfident_solution_edges;
vector<solution_edge_data> last_solution_edges;




struct pair_data
{
	IloNumVar pair_var;
	vertex_descriptor pair_source;
	vertex_descriptor pair_target;
	int shortest_distance;
	bool was_satisfie_last_round;
	bool was_satisfie_at_all;
	int last_distance_satisfied;
	int before_last_distance_satisfied;
	int first_distance_satisfied;
	bool stop_trying_to_satisfy;
};

struct protein_name_data
{
	vertex_descriptor protein;
	char*             name;
};


struct epw {
    epw(bidirected_basic_graph &g_) : g(g_) {}
    template <class Edge>
    void operator() (std::ostream &out, Edge e) 
	{
	if (hidden_map[e] != 0)
		{	
			out << "[color=blue]";
		}
	}
    
    bidirected_basic_graph &g;
};



protein_name_data protein_names_map[MAX_PROTEIN_ID];
protein_name_data filtered_names_map[MAX_PROTEIN_ID];
protein_index proteins_map[MAX_PROTEIN_ID + 1];

const int MAX_flow2 = 1;

int new_line = 0;
int max_path_between_pairs = 0;
int test_edges_directed_correct = 0;
int test_edges_directed_wrong = 0;
int num_of_ks = 0;
int num_of_pairs = 0;
int num_of_filetred_nodes = 0;
int num_of_proteins = 0;
edge_descriptor* parallel_edges_in_comp = NULL;


// Functions 
int  read_data			   (char** argv, IloModel model);			
int read_interaction(IloModel model,
					 char*  file_name,
					 int    max_id,
					 int   directed); //1-directed 2-undirected 3-complexes
void write_flow2_intersect(IloNumVar* pair_var, vertex_descriptor source, vertex_descriptor target, vertex_descriptor pair_source, vertex_descriptor pair_target, int shortest_path);
int compare(IloCplex cplex, IloModel model, int gap);
bool find_name(protein_name_data* protein_names_map, const char* protein_name, int max_id , int* out_mid);
bool prepare_ilp(IloModel model, bool allow_one_edge_path, IsoMap isoMap);
vertex_descriptor find_in_ordered_list(char* name_to_search);
void release();
void find_PATH_LENGTH_distances(bool*** distances, int size, bidirected_basic_graph* graph, vertex_myindex_map_t* index_map);
void fill_path(int start_location, int end_location, vector<vertex_descriptor>* path, vector<edge_descriptor>* path_edges,
			   bool*** distances_matrix,
			   ofstream* res_file,
			   std::vector<save_edge>*  save_edge_collection, bidirected_basic_graph* graph,  vertex_myindex_map_t* index_map);
void find_PATH_LENGTH_distances(bool*** distances, int size, bidirected_basic_graph* graph, vertex_myindex_map_t* index_map);

// Create a maximize objective
IloEnv env;
IloNumVarArray var(env);
IloObjective objective(env, 0.0, IloObjective::Maximize);

/* Keep a list of constraints to be kept in the model while expanding */
std::vector<IloInt> keep_constraints_Ids;

//IloNumVar* pairs_collection = NULL;
typedef std::vector<vertex_descriptor> targets;
map<vertex_descriptor,targets> sources;
std::vector <IloNumVar> rational_vars;
std::vector<pair_data> pairs;
std::vector<pair_data> copy_pairs;
std::vector<pair_data> expand_pairs;

double confidence_cutoff = 0.001;
int kpi=0,pdi=0,ppi=0;
int main( int argc, char** argv) 
{
	
	if (argc != 3)
	{
		cout << "Invalid arguments for preprocessing. needed 2 arguments: Input directory and output directory" << endl;
		return 1;
	}

	ofstream graphviz_input;
	ofstream graphviz_output;

	IloModel model(env);
	objective.setName("obj");
	keep_constraints_Ids.push_back(objective.getId());
	char* directory = new char[100];
	
	// Read data: proteins, directed interactions, undirected interactions, test set and pairs
	read_data(argv, model);

	// Create initial graph, can be viewed through dotty.exe
	strcpy(directory, argv[2]);
	graphviz_input.open (strcat(directory ,"initial_graph.txt"));
	write_graphviz(graphviz_input, bi_graph, make_label_writer(names_map), epw(solution_graph));
	graphviz_input.close();	


	// Prepare the ILP formula
	std::vector<vertex_descriptor> iso_vec(num_vertices(bi_graph));
	IsoMap isoMap(iso_vec.begin(), get(vertex_index, bi_graph));
	copy_graph(bi_graph, comp_graph, orig_to_copy(isoMap));
	IloCplex cplex(model);
	prepare_ilp(model, true, isoMap);	

	strcpy(directory, argv[2]);
	cplex.exportModel(strcat(directory, "shortest.lp"));

	// Solve the problem
	compare(cplex, model, 0);

	// Output directed graph
	strcpy(directory, argv[2]);
	graphviz_output.open (strcat(directory ,"solution_graph.txt"));
	write_graphviz(graphviz_output, bi_graph, make_label_writer(names_map), epw(solution_graph));
	graphviz_output.close();	
	
	return 0;
}


// This function is good for only one search, you call it each time with an alphabeticly higher name_to_search and searches for it in protein_names_map
vertex_descriptor find_in_ordered_list(char* name_to_search)
{
	static int current_id = 0;
	vertex_descriptor p = NULL;

	while (current_id < num_of_proteins)
	{
		if (strcmp(protein_names_map[current_id].name, name_to_search) == 0)
		{
			p = protein_names_map[current_id].protein;
			break;
		}
		current_id++;
	}
	return p;
}

// This is the new solution - using sources and targets, when one source has many targets, this solution should be faster
bool prepare_ilp(IloModel model, bool allow_one_edge_path, IsoMap isoMap)
{
	int count_improve = 0;
	int max_distance = 0;

	// List all shortest paths between a pair
	std::vector<pair_data> copy_pairs;
	const char* pair_name;
		
	// Create complementary graph
	reverse_graph<bidirected_basic_graph>comp_graph2(bi_graph);
	

	edge_iterator comp_edge, comp_edge_next, comp_edge_end;
	tie(comp_edge, comp_edge_end) = edges(comp_graph);
	for (comp_edge_next = comp_edge; comp_edge != comp_edge_end ; comp_edge = comp_edge_next) 
	{
		++comp_edge_next;
		remove_edge(*comp_edge, comp_graph);

	}
	for (tie (comp_edge, comp_edge_end) = edges(comp_graph2); comp_edge != comp_edge_end; comp_edge++)
	{

		vertex_descriptor vertex_source = source(*comp_edge, comp_graph2);
		vertex_descriptor vertex_target = target(*comp_edge, comp_graph2);
		pair<edge_descriptor, bool> a = add_edge(isoMap[vertex_source], isoMap[vertex_target], comp_graph);
		if (!a.second)
		{
			cout << "ERROR: edge was not created for comp_graph" << endl;
		}
	}




	// Running time improvement
	IloExpr sum_objective(env);

	int curr_pair = 0;

	map<vertex_descriptor,targets>::iterator pSource;
	for ( pSource=sources.begin() ; pSource != sources.end(); pSource++ )
	{
	
		// Init in shortest for current source in graph
		edge_iterator edge, edge_end;
		for (tie (edge, edge_end) = edges(bi_graph); edge != edge_end; edge++)
		{	
			ilp_improve_map[*edge] = false;
			in_shortest_map[*edge] = false;	
		}
		vertex_iterator vertex, vertex_end;
		for (tie (vertex, vertex_end) = vertices(bi_graph); vertex != vertex_end; vertex++)
		{	
			distances_map[*vertex] = 0;
		}

		// Init in shortest for current source in comp graph
		edge_iterator comp_edge, comp_edge_end;
		for (tie (comp_edge, comp_edge_end) = edges(comp_graph); comp_edge != comp_edge_end; comp_edge++)
		
		{
			in_shortest_map[*comp_edge] = false;
		}
		for (tie (vertex, vertex_end) = vertices(comp_graph); vertex != vertex_end; vertex++)
		{	
			distances_map[*vertex] = 0;
			comp_distances_map[*vertex] = 0;
		}

		vertex_descriptor pair_source = pSource->first;
		
		// boost::BFS from the source to all targets		
		breadth_first_search(bi_graph, pair_source, visitor(make_bfs_visitor(record_distances(distances_map, on_tree_edge()))));
		
		// Copy distances (distances_map will be changed by boost automatically next call to BFS)
		std::vector<int> d(num_vertices(bi_graph));
		//cout << "distances: " << endl;
		for (tie (vertex, vertex_end) = vertices(bi_graph); vertex != vertex_end; vertex++)
		{
			d[vertex_myindex_map[*vertex]] = distances_map[*vertex];			
		//	cout << names_map[pair_source] << " " << names_map[*vertex] << " " << distances_map[*vertex] << " " << d[vertex_myindex_map[*vertex]] << endl;
			distances_map[*vertex] = 0;
			if (max_distance < d[vertex_myindex_map[*vertex]])
			{
				max_distance = d[vertex_myindex_map[*vertex]];
			}
			
		}


		// Loop through all the pair's targets
		targets s_targets = pSource->second;
		targets s_targets_copy;
		while(!s_targets.empty())
		{

			vertex_descriptor curr_target_in_bi_graph = s_targets.back();
			vertex_descriptor curr_target = isoMap[curr_target_in_bi_graph];
			s_targets_copy.push_back(curr_target_in_bi_graph);
			s_targets.pop_back();

	
			int minimum_distance_per_pair = d[vertex_myindex_map[curr_target_in_bi_graph]];
		
			// If a path exists (path is bigger then zero, and bigger then one, if !allow_one_edge_path)
			// In case we want to limit path to bigger length (if one wants to expand over paths of length of 2 and more)
			if (minimum_distance_per_pair > 0) // path is bigger then zero,
			{
				// If the path cannot be of one edge, then remove this edge, compute distances, make sure that there is still a path and continue
				if (!allow_one_edge_path)
				{
					remove_edge(curr_target, isoMap[pair_source], comp_graph);
				}
				
				breadth_first_search(comp_graph, curr_target, visitor(make_bfs_visitor(record_distances(comp_distances_map, on_tree_edge()))));

				
				// Copy distances (distances_map will be changed by boost automatically next call to BFS)
				std::vector<int> comp_d(num_vertices(comp_graph));
				for (tie (vertex, vertex_end) = vertices(comp_graph); vertex != vertex_end; vertex++)
				{
					comp_d[vertex_comp_myindex_map[*vertex]] = comp_distances_map[*vertex];	
					distances_map[*vertex] = 0;
					comp_distances_map[*vertex] = 0;
				}

				minimum_distance_per_pair = comp_d[vertex_comp_myindex_map[pair_source]];
				if ((allow_one_edge_path) || (comp_d[vertex_comp_myindex_map[pair_source]] > 1)) // and bigger then one, if !allow_one_edge_path)
				{

					// Running time improvement
					IloExpr sum_flow2_per_pair(env);

					// Iterate over all edges and check their distance from pSource and from target,
					// to see if they participate in the shortest paths graph
					int distance_from_source;
					int distance_from_target;

					// Need to find the parallel edge to the edge we are examining. 
					// That is an edge whose target is this edge's source and vice versa. It distance is the complementart distance
					edge_descriptor parallel_edge;
					

					// Go over all edges and mark those who are in shortest
					edge_iterator searched, searched_end;
					for (tie (searched, searched_end) = edges(bi_graph); searched != searched_end; searched++)
					{


						vertex_descriptor edge_target = target(*searched,bi_graph);
						vertex_descriptor edge_source = source(*searched,bi_graph);
						
						distance_from_source =  d[vertex_myindex_map[edge_source]]; 
						distance_from_target =  comp_d[vertex_comp_myindex_map[isoMap[edge_target]]];

						// The distance from source and target needs to be equal to the shortest path, and that both source and
						// target were somehoe reached during the search
						if ((fabs(distance_from_source + distance_from_target + 1 - minimum_distance_per_pair) <  0.0001)
							&& (edge_source == pair_source || distance_from_source > 0) && (edge_target == curr_target_in_bi_graph || distance_from_target > 0))
						{
							in_shortest_map[*searched] = true;
							shortest_paths_counter_map[*searched] = shortest_paths_counter_map[*searched] + 1;

							if (ilp_improve_map[*searched] == false)
							{
								ilp_improve_map[*searched] = true;	

							}
							else
							{
								count_improve++;
							}

						}
					}
				}
				if (!allow_one_edge_path)
				{
					add_edge(curr_target,  isoMap[pair_source], comp_graph);
				}
			}
		
		}
		pSource->second = s_targets_copy;
					
		// Iterate all edges that are in shortest graph to create the constraint flow2_of_edge_for_pair <= orientation_of_edfe
		in_shortest_filter_edge_iterator in_shortest_edge, in_shortest_edge_end;
		for (tie (in_shortest_edge, in_shortest_edge_end) = edges(in_shortest_filtered); in_shortest_edge != in_shortest_edge_end; in_shortest_edge++)
		{
			
		
		
			// Add the flow2 variable f_SourcePair_TargetPair_SourceEdge_TargetEdge
			IloNumVar flow2_variable(env, 0, 1);
			rational_vars.push_back(flow2_variable);
			
			var.add(flow2_variable);

			// Save flow2 variable for later use - in the intersection constraints
			flows_map[*in_shortest_edge] = flow2_variable;

			IloConstraint c(flow2_variable <= MAX_flow2 * vars_map[*in_shortest_edge]);
			model.add(c);
		
		}

		// flow2 intersect - iterate on all in_shortest_edge edges, for each junction only if flow2 comes in flow2 can come out
		in_shortest_filter_vertex_iterator in_shortest_vertex, in_shortest_vertex_end;
		for (tie (in_shortest_vertex, in_shortest_vertex_end) = vertices(in_shortest_filtered); in_shortest_vertex != in_shortest_vertex_end; in_shortest_vertex++)
		{

			// write_flow2_intersect
			IloExpr sum_in_flow2(env);
			IloExpr sum_out_flow2(env);

			//// Sum in_edges
			bool any_in = false;
			in_shortest_filter_in_edge_iterator flow2_ei, flow2_ei_end;
			tie(flow2_ei, flow2_ei_end) = in_edges(*in_shortest_vertex, in_shortest_filtered);
			for (tie(flow2_ei, flow2_ei_end) = in_edges(*in_shortest_vertex, in_shortest_filtered); flow2_ei != flow2_ei_end ; flow2_ei++)
			{
				sum_in_flow2 += flows_map[*flow2_ei];
				any_in = true;
			}

			// Sum out_edges
			in_shortest_filter_out_edge_iterator flow2_eo, flow2_eo_end;
			tie(flow2_eo, flow2_eo_end) = out_edges(*in_shortest_vertex, in_shortest_filtered);
			for (tie(flow2_eo, flow2_eo_end) = out_edges(*in_shortest_vertex, in_shortest_filtered); flow2_eo != flow2_eo_end ; flow2_eo++)
			{
				if (any_in)
				{
					IloConstraint satisfy_flow2_constraint(flows_map[*flow2_eo] <= sum_in_flow2);
					model.add(satisfy_flow2_constraint);
				}
			}
			
			// If the current vertex is any target then the source,curr_target pair is satisfied if the in flow2 exists
			if (any_in)
			{

				// Possible run time improvement - change to map
				// Find the target, if any, that matches the current vertex 
				targets s_targets = pSource->second;
				targets s_targets_copy;
				vertex_descriptor curr_target = NULL;
				int curr_target_index = -1;
				int count_multiple_targets = 0;
				while(!s_targets.empty())
				{
					vertex_descriptor check_target = s_targets.back();
					s_targets_copy.push_back(check_target);
					s_targets.pop_back();

					if (check_target == *in_shortest_vertex)
					{
						curr_target = check_target;
						curr_target_index = vertex_myindex_map[curr_target];

						count_multiple_targets ++;
					}
				}
				pSource->second = s_targets_copy;



				if (curr_target != NULL)
				{

					for (int curr_pair_to_insert = 0; curr_pair_to_insert < count_multiple_targets; curr_pair_to_insert++)
					{

						pair_data pair_item;
						pair_item.pair_source = pair_source;
						pair_item.pair_target = curr_target;
						pair_item.was_satisfie_last_round = false;
						pair_item.was_satisfie_at_all = false;
						pair_item.shortest_distance = d[vertex_myindex_map[curr_target]];
						pair_item.last_distance_satisfied = 0;
						pair_item.before_last_distance_satisfied = 0;
						pair_item.first_distance_satisfied = 0;
						pair_item.stop_trying_to_satisfy = 0;


						// Create pair variable and use it to objective function (for the cplex to maximize)
						std::stringstream pair_itos;
						pair_itos << "s_" << names_map[pSource->first] << "_" << names_map[curr_target];
						pair_name = pair_itos.str().c_str();

						IloNumVar pair_var(objective(), 0, 1, IloNumVar::Float, pair_name);
						rational_vars.push_back(pair_var);

						// Save pair_var collection (to later change bounds)
						pair_item.pair_var = pair_var;


						IloConstraint satisfy_pair(pair_var <= sum_in_flow2);
						model.add(satisfy_pair);

						
						pairs.push_back(pair_item);
						curr_pair++;
					}
				}
			}
		}	
	}

	num_of_pairs = curr_pair;
	model.add(objective);

	return true;

}


bool find_name(protein_name_data* protein_names_map, const char* protein_name, int max_id , int* out_mid)
{
	int mid = max_id / 2;
	int prev_mid = 0;
	int temp = 0;
	int search_span = max_id;
	int comp = strcmp(protein_name, protein_names_map[mid].name);
	bool found = false;


	// Look for first and second protein
	while (!found && (search_span >= 0))
	{
		double help = (prev_mid - mid > 0 ? prev_mid - mid : mid - prev_mid);
		search_span = ceil(help / 2);
		if (search_span > mid )
		{
			search_span = 1;
		}

		temp = prev_mid;
		prev_mid = mid;
		if (comp == 0)
		{
			found = true;
		}
		else if (comp > 0)
		{
			mid = mid + search_span;

			
			if (mid > max_id)
			{
				mid = max_id - 1;
				if (search_span < 1)
				{
					search_span = -1;
				}
			}
			else if (mid < 0)
			{
				mid = 0;
				search_span = -1;
			}
			if (mid == temp)
			{
				search_span = -1;
			}	
				

		}
		else
		{

			mid = mid - search_span;
			
			if (mid > max_id)
			{
				mid = max_id - 1;
				if (search_span < 1)
				{
					search_span = -1;
				}
			}
			else if (mid < 0)
			{
				mid = 0;
				search_span = -1;
			}
			if (mid == temp) //Why do we need this?
			{
				search_span = -1;
			}	
		}

		comp = strcmp(protein_name, protein_names_map[mid].name);
	}
	*out_mid = mid;
	return found;
}


int read_interaction(IloModel model,
					 char*  file_name,
					 int    max_id,
					 int   directed) //1-directed 2-undirected 3-complexes
{
	ifstream ppi_file;
	ppi_file.open(file_name);

	int num_interactions_read = 0;
	if (ppi_file.good())
	{

		// Read interactions until end of file or sufficient data	
		char* first_protein_name = new char[20];
		char* second_protein_name = new char[20];
		char* weight_string = new char[20];
		double weight;
	
		while (!ppi_file.eof())
		{
			
			vertex_descriptor v1 = NULL;
			vertex_descriptor v2 = NULL;
			ppi_file.getline(first_protein_name, 100, ' ');

			// And for second protein
			ppi_file.getline(second_protein_name, 100, ' ');
			ppi_file.getline(weight_string, 100, '\n');
			weight = atof(weight_string);

			
			if (!ppi_file.eof())
			{
				int* mid = new int;
				bool found1 = find_name(protein_names_map, first_protein_name, max_id, mid);
				
				if (found1)
				{
					v1 = protein_names_map[*mid].protein;
				}
				else if (directed != 3)
				{
					cout << "ERROR: protein " << first_protein_name << " not found in the list (in file = " << directed << ")" << endl;
				}

				bool found2 = find_name(protein_names_map, second_protein_name, max_id, mid);
				
				if (found2)
				{
					v2 = protein_names_map[*mid].protein;
				}
				else if (directed != 3)
				{
					cout << "ERROR: protein " << second_protein_name << " not found in the list (in file = " << directed << ")" << endl;
				}
							
				
				// Create an edge accordingly
				bool take = true;
				edge_descriptor ret_edge;

				// Filter out edges that belong to complexes
				if (directed == 3) //complexes
				{
					take = false;
					if (found1 && found2)
					{
						bool bRet; 
						edge_descriptor e; 
						tie(e,bRet) = edge(v1, v2, bi_graph);
						if (bRet == true)
						{
							remove_edge(e, bi_graph);
						}
						tie(e,bRet) = edge(v2, v1, bi_graph);
						if (bRet == true)
						{
							remove_edge(e, bi_graph);
						}

						
					}
				}

				
				if (v1 == v2)
				{
					take = false;
	
				}
				if (take)
				{
					num_interactions_read++;

					// Check if such an edge already exists (We preffer to keep test edges over all else, and keep directed edges over undirected)
					bool bRet, bOppositeRet; 
					edge_descriptor e; 
					tie(e,bRet) = edge(v1, v2, bi_graph);

					// if we are adding undirected edges at the moment, then we should look for other direction as well, and if it exists then not add it as well (because the edge
					// was alreadt added as directed or test
					if (directed == 2) //undirected
					{
						tie(e,bOppositeRet) = edge(v2, v1, bi_graph);
					}
					else
					{
						bOppositeRet = false;
					}

					if (!bRet && !bOppositeRet)
					{
						

						pair<edge_descriptor, bool> a = add_edge( v1, v2, bi_graph);  
						weightmap[a.first] = 1;
						in_shortest_map[a.first] = false; 
						shortest_paths_counter_map[a.first] = 0;
						test_map[a.first] = 0;
						hidden_map[a.first] = 0; //not a test edge
						directions_map[a.first] = 1;
						parallel_confidence_map[a.first] = weight;

						// Create the cplex variables
						std::stringstream itos;
						itos << "o_" << names_map[v1] << "_" << names_map[v2];
						const char* var_name = itos.str().c_str();
						IloNumVar edge_variable(env, 0, 1, ILOBOOL, var_name);
						vars_map[a.first] = edge_variable;



						// If it's a ppi (undirected) we should add also a directed edge in the oposite direction
						if (directed == 2) //undirected
						{

							directions_map[a.first] = 0;
							pair<edge_descriptor, bool> a2 = add_edge( v2, v1, bi_graph);  
							weightmap[a2.first] = 1;
							in_shortest_map[a2.first] = false; 
							shortest_paths_counter_map[a2.first] = 0;
							test_map[a2.first] = 0;
							hidden_map[a2.first] = 0; // not a test edge
							directions_map[a2.first] = 0;
							parallel_confidence_map[a2.first] = weight;

							// Create the cplex variables
							itos.str("");
							itos << "o_" << names_map[v2] << "_" << names_map[v1];
							var_name = itos.str().c_str();
							IloNumVar opposite_edge_variable(env, 0, 1, ILOBOOL, var_name);
							vars_map[a2.first] = opposite_edge_variable;

							IloExpr sum_of_directions_in_edge(env);
							sum_of_directions_in_edge = edge_variable + opposite_edge_variable;
							IloConstraint only_one_direction_constraint(sum_of_directions_in_edge <= 1);
							model.add(only_one_direction_constraint);
							keep_constraints_Ids.push_back(only_one_direction_constraint.getId());

							constraints_map[a.first] = only_one_direction_constraint;// Save so we can check shadow price later
							constraints_map[a2.first] = only_one_direction_constraint;// Save so we can check shadow price later
						}
						else
						{
							IloConstraint only_one_direction_constraint(edge_variable <= 1);
							model.add(only_one_direction_constraint);

						
						}
					}
				}
				delete mid;
			}
		}

		ppi_file.close();
	}
	else
	{
		cout << "ERROR: in file " << file_name << " not opened\n";	
	}
	
	return num_interactions_read;
}

// int read_data()
//================
// Read from file vertices (name, color) edges (source, target, direction)
int read_data(char** argv, IloModel model)
{
	
	// Read proteins and put in the list
	ifstream sorted_protein_file;
	int current_id = 0;
	char* input_directory = new char[60];
	char* output_directory = new char[60];
	
	strcpy(input_directory, argv[1]);
	sorted_protein_file.open(strcat(input_directory, "names.txt"));
	if (sorted_protein_file.good())
	{
		vertex_descriptor v1;
		
		char* dummy = new char[100];
		char* itoa_helper = new char[20];
		while (!sorted_protein_file.eof())
		{
		
			protein_names_map[current_id].name = new char[20];
			sorted_protein_file.getline(protein_names_map[current_id].name, 100, '\n');

			v1 = add_vertex(vertex_properties(protein_names_map[current_id].name), bi_graph);
		
	
		
			protein_names_map[current_id].protein = v1;
		
			current_id++;
			if (current_id > MAX_PROTEIN_ID)
			{
				cout << "current id exceeded MAX_PROTEIN_ID. MAX_PROTEIN_ID = " << MAX_PROTEIN_ID << endl;
				return 1;
			}
			
		}
		
		delete [] dummy;
		delete [] itoa_helper;
		sorted_protein_file.close();
	}
	current_id--;
	num_of_proteins = current_id;
	
	// Sort the array
	int i, j, flag = 1;    // set flag to 1 to start first pass
	protein_name_data temp;             // holding variable
	for(i = 1; i <= current_id && flag; i++)
	{
	  flag = 0;
	  for (j=0; j < (current_id - 1); j++)
	 {
		  if (strcmp(protein_names_map[j + 1].name, protein_names_map[j].name) < 0)// ascending order simply changes to <
		  { 
			  
				temp = protein_names_map[j];             // swap elements
				protein_names_map[j] = protein_names_map[j + 1];
				protein_names_map[j + 1] = temp;
				flag = 1;               // indicates that a swap occurred.
		   }
	  }
	}


	num_of_ks = 0;
	ifstream test_file;
	strcpy(input_directory, argv[1]);
	sorted_protein_file.open(strcat(input_directory, "test"));
	test_file.open (input_directory); 
	
	
	if (test_file.good())
	{

		char* first_protein_name = new char[20];
		char* second_protein_name = new char[20];
		char* weight_string = new char[20];
		double weight;
		int i = 0;
		while (!test_file.eof())
		{
			
			int max_id = current_id;
			test_file.getline(first_protein_name, 100, ' ');

			// And for second protein
			test_file.getline(second_protein_name, 100, ' ');	
			test_file.getline(weight_string, 100, '\n');	
			weight = atof(weight_string);

		
			if (!test_file.eof())
			{

				i++;
						
				int* mid = new int;
				bool first_found = find_name(protein_names_map, first_protein_name, max_id, mid);
				
				vertex_descriptor v1 = NULL;
				if (first_found) 
				{
					v1 = protein_names_map[*mid].protein;
				}
				else
				{
					cout << "ERROR: protein " << first_protein_name << " not found in the list " << endl;
				}

				bool second_found = find_name(protein_names_map, second_protein_name, max_id, mid);
				vertex_descriptor v2 = NULL;
				if (second_found)
				{
					v2 = protein_names_map[*mid].protein;
				}
				else
				{
					cout << "ERROR: protein " << second_protein_name << " not found in the list " << endl;
				}
				if (first_found && second_found)
				{

					edge_descriptor ret_edge;
					bool take = true;
					if (v1 == v2)
					{
						take = false;
						cout << "Notice: self loop accured" << endl;
					}
					if (take)
					{
						bool bRet; 
						edge_descriptor e; 
						tie(e,bRet) = edge(v1, v2, bi_graph);
						
						if (bRet)
						{
							cout << "ERROR: duplicated test edge " << names_map[v1] << " " << names_map[v2] << endl;
						}
						else
						{

					
							// Create the 2 directed edges accordingly

							// First edge is the correct oriented edge - this is the test
							pair<edge_descriptor, bool> a = add_edge( v1, v2, bi_graph);  
							weightmap[a.first] = 1;
							directions_map[a.first] = 0;
							test_map[a.first] = 1;
							hidden_map[a.first] = 1; // test edge correct orientation
							in_shortest_map[a.first] = false;
							shortest_paths_counter_map[a.first] = 0;
							parallel_confidence_map[a.first] = weight;
							
							// Second edge - only to hide the direction
							pair<edge_descriptor, bool> a2 = add_edge( v2, v1, bi_graph);
							weightmap[a2.first] = 1;
							directions_map[a2.first] = 0;
							test_map[a2.first] = 0;
							hidden_map[a2.first] = 2; // test edge incorrect orientation
							in_shortest_map[a2.first] = false;
							shortest_paths_counter_map[a2.first] = 0;
							parallel_confidence_map[a2.first] = weight;
							
							// Create the cplex variables
							std::stringstream itos;
							itos << "o_" << names_map[v1] << "_" << names_map[v2];
							const char* var_name = itos.str().c_str();
							IloNumVar edge_variable(env, 0, 1, ILOBOOL, var_name);
							vars_map[a.first] = edge_variable;

							itos.str("");
							itos << "o_" << names_map[v2] << "_" << names_map[v1];
							var_name = itos.str().c_str();
							IloNumVar opposite_edge_variable(env, 0, 1, ILOBOOL, var_name);
							vars_map[a2.first] = opposite_edge_variable;
							
							IloExpr sum_of_directions_in_edge(env);
							sum_of_directions_in_edge = edge_variable + opposite_edge_variable;
							IloConstraint only_one_direction_constraint(sum_of_directions_in_edge <= 1);
							only_one_direction_constraint.setName("keep");
							model.add(only_one_direction_constraint);
							keep_constraints_Ids.push_back(only_one_direction_constraint.getId());
							
							constraints_map[a.first] = only_one_direction_constraint;// Save so we can check shadow price later
							constraints_map[a2.first] = only_one_direction_constraint;// Save so we can check shadow price later

							pdi++;
						}
					}
				}
				delete mid;
			}
		}
		delete (first_protein_name);
		delete (second_protein_name);
		delete (weight_string);

		test_file.close();
	}

	strcpy(input_directory, argv[1]);
	strcat(input_directory, "directed");
	read_interaction(model, input_directory, current_id, 1); //1-directed
	strcpy(input_directory, argv[1]);
	strcat(input_directory, "undirected");
	read_interaction(model, input_directory, current_id, 2); //2-undirected

	
	ifstream ko_file; 
	strcpy(input_directory, argv[1]);
	ko_file.open (strcat(input_directory, "ko")); 

	num_of_pairs = 0;
	int num_of_potential_pairs = 0;
	map<vertex_descriptor,targets>::iterator map_it;
	if (ko_file.good())
	{

		// Read interactions until end of file or sufficient data
		char* first_protein_name = new char[20];
		char* second_protein_name = new char[20];
		while (!ko_file.eof())
		{
		
			ko_file.getline(first_protein_name, 100, ' ');
			// Get type
			// And for second protein
			ko_file.getline(second_protein_name, 100, ' '); 

			if (!ko_file.eof())
			{

				int* mid = new int;
				bool found_first = find_name(protein_names_map, first_protein_name, current_id, mid);
				vertex_descriptor v1 = protein_names_map[*mid].protein;
				bool found_second = find_name(protein_names_map, second_protein_name, current_id, mid);
				vertex_descriptor v2 = protein_names_map[*mid].protein;

				num_of_potential_pairs++;

				// If one wasn't found produce an error
				if (!found_first || !found_second)
				{
					cout << "ERROR : a knouckout pair with unrecognized proteins: " << first_protein_name << "ko" << second_protein_name << endl;
				}
				// Create a pair
				else
				{
					num_of_pairs++;

					// Prepare a list of sources, to each source a collection of its targets
					map_it = sources.find(v1);

					// The particular source was already inserted to the collection
					if (map_it !=  sources.end())
					{
						// Insert the next target
						(map_it->second).push_back(v2);
					}
					// First time for cuurent source
					else
					{
						targets first_target;
						first_target.push_back(v2);
						sources.insert(pair<vertex_descriptor,targets>(v1,first_target));
					}
				}
				delete mid;
			}
			// Read rest of line
			ko_file.getline(first_protein_name, 100, '\n');
			

		}
		delete (first_protein_name);
		delete (second_protein_name);
		ko_file.close();
	}

	else
	{
		cout << "ERROR: knockout file not opened\n";	
	}
	
	int ver_index, edge_index = 0;
	edge_iterator edge, edge_end;
	for (tie (edge, edge_end) = edges(bi_graph); edge != edge_end; edge++)
	{
	

		// Create indexes
		myindex_map[*edge]= edge_index;
		edge_index++;

	}

	vertex_iterator vertex, vertex_end;
	for (tie (vertex, vertex_end) = vertices(bi_graph); vertex != vertex_end; vertex++)
	{

		// Create indexes
		vertex_myindex_map[*vertex]= ver_index;
		ver_index++;

	}

	delete (input_directory);
	delete (output_directory);
	
	return num_of_ks;
}

int compare(IloCplex cplex, IloModel model, int gap)
{
	ofstream res_file;
	int pairs_satisfied = -1;
	int num_of_changes = 0;

	edge_iterator curr_edge, edge_end, edge_next;

	cplex.setParam(IloCplex::RootAlg, IloCplex::Network); // Optimize solution using flow charts
	cplex.setParam(IloCplex::VarSel, CPX_VARSEL_STRONG); // Strong branching (3) causes variable selection based on partially solving a number of subproblems with tentative branches to see which branch is the most promising. This strategy can be effective on large, difficult MIP problems
	cplex.setParam(IloCplex::WorkMem, 1280); // Working memory, should be up to half of available RAM
	cplex.setParam(IloCplex::DiveType, 3); // Guided dive, recommanded in cases of many boolean integers variables
	
	double cplex_gap = (double)gap/100;
	cplex.setParam(IloCplex::EpGap, cplex_gap); // (cplex default is 1%)

	// Set priority: Orientation binary variables are the only ones that get priority (others are automaticly assigned 0), and they get the priority accoding to how many shortest paths they participate in.
	for (tie (curr_edge, edge_end) = edges(bi_graph); curr_edge != edge_end; curr_edge++)
	{
		if ((shortest_paths_counter_map[*curr_edge] > 0) && (test_map[*curr_edge] != 22))
		{
			cplex.setPriority(vars_map[*curr_edge], shortest_paths_counter_map[*curr_edge]);
		}
	}
	
	if (!cplex.solve()) 
	{
		 cout << "Failed to optimize LP" << endl;
		 throw(-1);
	}


	pairs_satisfied = cplex.getObjValue();
	cout << "Solution status = " << cplex.getStatus() << endl;
	cout << "Solution value  = " << cplex.getObjValue() << endl;

	
	// Read all pairs and keep those which were not satisfied
	std::vector<pair_data> copy_pairs_collection;


	cout << "cplex solver finished" << endl;
		

	// Check the orientation result for the test edges		
	int curr_edge_examined = 0;
	edge_descriptor previous_edge;
		
	for (tie (curr_edge, edge_end) = edges(bi_graph); curr_edge != edge_end; curr_edge++)
	{
		if (hidden_map[*curr_edge] != 0)
		{

		
			IloNumVar currVar = vars_map[*curr_edge];
			int val = cplex.getValue(currVar);	

			// Keep this direction, mark parallel edge as to be removed
			if (val == 1)
			{

				// Mark edge as to be kept, remove the opposite direction
				edge_descriptor opposite_edge;
				bool bRet = false;
				tie(opposite_edge,bRet) = edge(target(*curr_edge, bi_graph), source(*curr_edge, bi_graph), bi_graph);
				if (bRet)
				{
					test_map[opposite_edge] = 3;
				}
				
			}
			else if (val != 0 && val != 1) 
			{
				cout << "ERROR - Binary variable got a very unbinary value" << endl;
			}

			curr_edge_examined++;
		}
		
	}

	// Remove all edges that were hidden as test edges and were oriented to the other direction
	tie (curr_edge, edge_end) = edges(bi_graph);
	for (edge_next = curr_edge; curr_edge != edge_end; curr_edge = edge_next)
	{
		edge_next++;

		if ((test_map[*curr_edge] == 3) && (hidden_map[*curr_edge] != 0))
		{
		
			remove_edge(*curr_edge, bi_graph);
		}
	}

	return num_of_changes;
}